"""Tests for epoch.py"""

import numpy as np
import pandas as pd
import pytest
import xarray as xr
from timeflux.core.branch import Branch
from timeflux.core.exceptions import WorkerInterrupt
from timeflux.helpers import testing as helpers
from timeflux.nodes.epoch import Epoch, Samples

def test_samples_no_rate():
    # Trigger an error if not rate can be found
    node = Samples(trigger='test', length=0.6, buffer=5)
    node.i.data = helpers.DummyData().next(20)
    with pytest.raises(WorkerInterrupt):
        node.update()

def test_samples_rate_from_meta():
    # Get rate from metadata
    node = Samples(trigger='test', length=0.6, buffer=5)
    node.i.data = helpers.DummyData().next(20)
    node.i.meta = { "rate": 250 }
    node.update()
    assert node._rate == 250

def test_samples_buffer():
    # Make sure that the circular buffer always keeps a little bit of past data
    node = Samples(trigger='test', length=0.6, buffer=3, rate=10)
    data = helpers.DummyData(rate=10)
    node.i.data = data.next(20)
    node.update()
    assert len(node._buffer) == 20
    node.i.data = data.next(20)
    node.update()
    length = round((node._buffer.index[-1] - node._buffer.index[0]).total_seconds())
    assert length == 3
    assert len(node._buffer) == 30
    node.i.data = data.next(20)
    node.update()
    length = round((node._buffer.index[-1] - node._buffer.index[0]).total_seconds())
    assert length == 3
    assert len(node._buffer) == 30

def test_samples_no_epoch():
    # Send enough data for an epoch, but no event
    node = Samples(trigger='test', length=0.6, buffer=5, rate=10)
    node.i.data = helpers.DummyData().next(20)
    node.update()
    assert node.o.data == None

def test_samples_invalid_event():
    # Make sure no epoch is fetched on an invalid trigger
    node = Samples(trigger='test', length=0.6, buffer=5, rate=10)
    node.i.data = helpers.DummyData().next(20)
    time = node.i.data.index[1]  # Sync event to second sample
    event = pd.DataFrame([['invalid', 'foobar']], [time], columns=['label', 'data'])  # Generate a trigger event
    node.i_events.data = event
    node.update()
    assert node.o.data == None

def test_samples_outdated_event(caplog):
    # Send an event trigger in the past
    node = Samples(trigger='test', length=0.6, buffer=5, rate=10)
    node.i.data = helpers.DummyData().next(5)
    time = pd.Timestamp('2010-01-01 00:00:00.000000000')
    event = pd.DataFrame([['test', 'foobar']], [time], columns=['label', 'data'])  # Generate a trigger event
    node.i_events.data = event
    node.update()
    assert node.o.data == None
    assert caplog.record_tuples[0][2].startswith('Oudated event')

def test_samples_send_event_short_data():
    # Send an event trigger, but not enough data to capture the epoch in one round
    node = Samples(trigger='test', length=0.6, buffer=5, rate=10)
    node.i.data = helpers.DummyData().next(5)
    time = node.i.data.index[0]  # Sync event to first sample
    event = pd.DataFrame([['test', 'foobar']], [time], columns=['label', 'data'])  # Generate a trigger event
    node.i_events.data = event
    node.update()
    assert node.o.data == None

def test_samples_receive_epoch():
    # Send enough data to complete one epoch in one one round
    node = Samples(trigger='test', length=0.6, buffer=5, rate=10)
    node.i.data = helpers.DummyData().next(20)
    time = node.i.data.index[1]  # Sync event to second sample
    event = pd.DataFrame([['test', 'foobar']], [time], columns=['label', 'data'])  # Generate a trigger event
    node.i_events.data = event
    node.update()
    expected_meta = {
        'rate': 10,
        'epoch': {
            'onset': pd.Timestamp('2018-01-01 00:00:00.104507143'),
            'context': 'foobar'
        }
    }
    assert node.o == node.o_0
    assert node.o.meta == expected_meta
    assert node.o.data.index[0] == expected_meta["epoch"]["onset"]
    assert len(node.o.data) == 6

def test_samples_append_epoch():
    # Complete one epoch in two passes
    node = Samples(trigger='test', length=0.6, buffer=5, rate=10)
    data = helpers.DummyData()
    node.i.data = data.next(5)
    time = node.i.data.index[1]  # Sync event to second sample
    event = pd.DataFrame([['test', 'foobar']], [time], columns=['label', 'data'])  # Generate a trigger event
    node.i_events.data = event
    node.update()
    assert len(node._epochs[0]["data"]) == 4
    assert node.o.data == None
    node.clear()
    node.i.data = data.next(5)
    node.update()
    expected_meta = {
        'rate': 10,
        'epoch': {
            'onset': pd.Timestamp('2018-01-01 00:00:00.104507143'),
            'context': 'foobar'
        }
    }
    assert node.o.meta == expected_meta
    assert node.o.data.index[0] == expected_meta["epoch"]["onset"]
    assert len(node.o.data) == 6

def test_samples_non_monotonic_buffer():
    # Make sure that the buffer does not error on non-monotonic data
    node = Samples(trigger='test', length=0.6, buffer=5, rate=10)
    node.i.data = helpers.DummyData().next()
    node.update()
    node.i.data = helpers.DummyData().next()
    node.update()

def test_samples_non_monotonic_epoch():
    # Handle epoching on non-monotonic data
    node = Samples(trigger='test', length=0.6, buffer=5, rate=10)
    data = helpers.DummyData()
    data_past = data.next(5)
    data_present = data.next(5)
    data_future = data.next(5)
    node.i.data = data_present
    time = data_present.index[1]
    event = pd.DataFrame([['test', 'foobar']], [time], columns=['label', 'data'])
    node.i_events.data = event
    node.update()
    assert len(node._epochs[0]["data"]) == 4
    assert node.o.data == None
    node.clear()
    node.i.data = data_past
    node.update()
    assert len(node._epochs[0]["data"]) == 4
    node.clear()
    node.i.data = data_future
    node.update()
    expected_meta = {
        'rate': 10,
        'epoch': {
            'onset': pd.Timestamp('2018-01-01 00:00:00.595580836'),
            'context': 'foobar'
        }
    }
    assert node.o.meta == expected_meta
    assert node.o.data.index[0] == expected_meta["epoch"]["onset"]
    assert len(node.o.data) == 6

def test_samples_future_event():
    # Handle future events
    node = Samples(trigger='test', length=0.6, buffer=5, rate=10)
    node.i.data = helpers.DummyData().next(10)
    time = pd.Timestamp('2042-01-01')
    event = pd.DataFrame([['test', 'foobar']], [time], columns=['label', 'data'])
    node.i_events.data = event
    node.update()
    assert len(node._epochs) == 1
    assert node._epochs[0]["data"] == None
    assert node.o.data == None

def test_samples_multiple_events():
    # Make sure the correct onset is caught
    node = Samples(trigger='test', length=0.6, buffer=5, rate=10)
    node.i.data = helpers.DummyData().next(10)
    event = pd.DataFrame(
        [
            ['invalid_1', 'foo'],
            ['test', 'bar'],
            ['invalid_2', 'baz']
        ], [
            node.i.data.index[0],
            node.i.data.index[1],
            node.i.data.index[2]
        ], columns=['label', 'data'])
    node.i_events.data = event
    node.update()
    assert node.o.meta['epoch']['onset'] == node.i.data.index[1]

def test_samples_multiple_triggers():
    # Mutliple epochs should be returned
    node = Samples(trigger='test', length=0.6, buffer=5, rate=10)
    data = helpers.DummyData()
    data_1 = data.next(10)
    data_2 = data.next(3)
    data_3 = data.next(10)
    event = pd.DataFrame(
        [
            ['test', 'foo'],
            ['test', 'bar'],
            ['test', 'baz']
        ], [
            data_1.index[0],
            data_1.index[8],
            data_2.index[0]
        ], columns=['label', 'data'])
    node.i_events.data = event
    node.i.data = data_1
    node.update()
    assert len(node._epochs) == 2
    assert node.o.meta['epoch']['context'] == 'foo'
    assert len(node._epochs[0]['data']) == 2
    assert node._epochs[1]['data'] == None
    node.clear()
    node.i.data = data_2
    node.update()
    assert len(node._epochs) == 2
    assert len(node._epochs[0]['data']) == 5
    assert len(node._epochs[1]['data']) == 3
    node.clear()
    node.i.data = data_3
    node.update()
    assert len(node._epochs) == 0
    assert node.o_0.meta['epoch']['context'] == 'bar'
    assert node.o_1.meta['epoch']['context'] == 'baz'

def test_samples_jumbled_events():
    # Send correct epochs
    node = Samples(trigger='test', length=0.6, buffer=5, rate=10)
    data = helpers.DummyData()
    data_1 = data.next(10)
    data_2 = data.next(10)
    event = pd.DataFrame(
        [
            ['test', 'foo'],
            ['test', 'bar'],
        ], [
            data_2.index[0],
            data_1.index[0]
        ], columns=['label', 'data'])
    node.i_events.data = event
    node.i.data = data_1
    node.update()
    assert node.o.meta['epoch']['context'] == 'bar'
    node.clear()
    node.i.data = data_2
    node.update()
    assert node.o.meta['epoch']['context'] == 'foo'

def test_samples_unsynced_event():
    # Correctly triggers on the immediately following timestamp
    node = Samples(trigger='test', length=0.6, buffer=5, rate=10)
    node.i.data = helpers.DummyData().next(10)
    time = pd.Timestamp('2018-01-01')
    event = pd.DataFrame([['test', 'foobar']], [time], columns=['label', 'data'])
    node.i_events.data = event
    node.update()
    assert node.o.meta['epoch']['onset'] == time
    assert node.o.data.index[0] == node.i.data.index[1]

def test_samples_offset_positive():
    # Apply a positive offset to the signal
    node = Samples(trigger='test', length=0.3, buffer=5, rate=10, offset=0.2)
    node.i.data = helpers.DummyData().next(10)
    time = pd.Timestamp('2018-01-01')
    event = pd.DataFrame([['test', 'foobar']], [time], columns=['label', 'data'])
    node.i_events.data = event
    node.update()
    assert node.o.meta['epoch']['onset'] == time
    assert node.o.data.index[0] == node.i.data.index[2]

def test_samples_offset_negative():
    # Apply a negative offset to the signal
    node = Samples(trigger='test', length=0.3, buffer=5, rate=10, offset=-0.2)
    node.i.data = helpers.DummyData().next(10)
    time = pd.Timestamp('2018-01-01 00:00:00.5')
    event = pd.DataFrame([['test', 'foobar']], [time], columns=['label', 'data'])
    node.i_events.data = event
    node.update()
    assert node.o.meta['epoch']['onset'] == time
    assert node.o.data.index[0] == node.i.data.index[3]


data = helpers.DummyData()

rate = 10
before = .2
after = .6
node = Epoch(event_trigger='test', before=before, after=after)


def test_no_epoch():
    # Send enough data for an epoch, but no event
    node.i.data = data.next(20)
    node.update()
    assert node.o.data == None


def test_send_event_short_data():
    # Send an event trigger, but not enough data to capture the epoch in one round
    node.clear()
    node.i.data = data.next(5)
    time = node.i.data.index[1]  # Sync event to second sample
    event = pd.DataFrame([['test', 'foobar']], [time], columns=['label', 'data'])  # Generate a trigger event
    node.i_events.data = event
    node.update()
    assert node.o.data == None


def test_receive_epoch():
    # Send enough data to complete an epoch
    node.clear()
    node.i.data = data.next()
    node.update()
    expected_meta = {
        'epoch': {
            'onset': pd.Timestamp('2018-01-01 00:00:02.096394939'),
            'context': 'foobar',
            'before': before,
            'after': after
        }
    }
    expected_data = pd.DataFrame(
        [
            [0.176528, 0.220486, 0.186438, 0.779584, 0.350125],
            [0.057843, 0.969103, 0.883786, 0.927752, 0.994908],
            [0.173895, 0.396242, 0.758238, 0.696021, 0.153896],
            [0.815833, 0.224441, 0.223818, 0.536974, 0.592940],
            [0.580086, 0.091487, 0.877461, 0.265600, 0.129515],
            [0.888748, 0.955651, 0.862128, 0.809516, 0.655242],
            [0.550857, 0.086987, 0.408453, 0.372689, 0.259754],
            [0.723420, 0.495876, 0.081046, 0.220183, 0.683259]
        ],
        [
            pd.Timestamp('2018-01-01 00:00:01.897912291'),
            pd.Timestamp('2018-01-01 00:00:02.001118529'),
            pd.Timestamp('2018-01-01 00:00:02.096394939'),
            pd.Timestamp('2018-01-01 00:00:02.197921447'),
            pd.Timestamp('2018-01-01 00:00:02.298663618'),
            pd.Timestamp('2018-01-01 00:00:02.399560700'),
            pd.Timestamp('2018-01-01 00:00:02.502851760'),
            pd.Timestamp('2018-01-01 00:00:02.596996738')
        ]
    )
    assert node.o.meta == expected_meta
    pd.testing.assert_frame_equal(node.o.data, expected_data)


def test_invalid_event():
    # Make sure no epoch is fetched on an invalid trigger
    node.clear()
    node.i.data = data.next()
    time = node.i.data.index[0]
    event = pd.DataFrame([['invalid', 'foobar']], [time], columns=['label', 'data'])
    node.i_events.data = event
    node.update()
    assert node.o.data == None


def test_multiple_events():
    # Make sure the correct onset is caught
    node.clear()
    node.i.data = data.next()
    event = pd.DataFrame(
        [
            ['invalid_1', 'foo'],
            ['test', 'bar'],
            ['invalid_2', 'baz']
        ], [
            node.i.data.index[0],
            node.i.data.index[1],
            node.i.data.index[2]
        ], columns=['label', 'data'])
    node.i_events.data = event
    node.update()
    assert node.o.meta['epoch']['onset'] == pd.Timestamp('2018-01-01 00:00:04.598117111')


def test_multiple_triggers():
    # Mutliple epochs should be returned
    node.clear()
    node.i.data = data.next()
    event = pd.DataFrame(
        [
            ['test', 'foo'],
            ['test', 'bar'],
        ], [
            node.i.data.index[0],
            node.i.data.index[1],
        ], columns=['label', 'data'])
    node.i_events.data = event
    node.update()
    assert node.o_0.meta['epoch']['context'] == 'foo'
    assert node.o_1.meta['epoch']['context'] == 'bar'


def test_unsynced_event():
    # The epoch must be fetched even if the trigger time does not match exactly
    node.clear()
    node.i.data = data.next(20)
    time = pd.Timestamp('2018-01-01 00:00:07.100')
    event = pd.DataFrame([['test', 'foobar']], [time], columns=['label', 'data'])
    node.i_events.data = event
    node.update()
    expected_indices = pd.DatetimeIndex([
        pd.Timestamp('2018-01-01T00:00:06.904868869'),
        pd.Timestamp('2018-01-01T00:00:07.002722448'),
        pd.Timestamp('2018-01-01T00:00:07.096987157'),
        pd.Timestamp('2018-01-01T00:00:07.195055221'),
        pd.Timestamp('2018-01-01T00:00:07.303154614'),
        pd.Timestamp('2018-01-01T00:00:07.402068573'),
        pd.Timestamp('2018-01-01T00:00:07.502290072'),
        pd.Timestamp('2018-01-01T00:00:07.602712703'),
        pd.Timestamp('2018-01-01T00:00:07.695740447')
    ])
    pd.testing.assert_index_equal(node.o.data.index, expected_indices)


def test_receives_event_before_data():
    # receives event before data
    # NB: this use case can also happen if before<0
    before = 0
    data.reset()
    node = Epoch(event_trigger='test', before=before, after=.6)
    node.clear()
    node.i.data = data.next(5)
    time = data._data.index[6]  # Sync event right after last sample
    event = pd.DataFrame([['test', 'foobar']], [time], columns=['label', 'data'])  # Generate a trigger event
    node.i_events.data = event
    node.update()
    node.i_events.data = None
    node.i.data = data.next(10)
    node.update()
    expected_data = pd.DataFrame(
        [
            [0.037008, 0.59627, 0.230009, 0.120567, 0.076953],
            [0.696289, 0.339875, 0.724767, 0.065356, 0.31529],
            [0.539491, 0.790723, 0.318753, 0.625891, 0.885978],
            [0.615863, 0.232959, 0.024401, 0.870099, 0.021269],
            [0.874702, 0.528937, 0.939068, 0.798783, 0.997934],
            [0.350712, 0.767188, 0.401931, 0.479876, 0.627505]
        ],
        [
            pd.Timestamp('2018-01-01 00:00:00.595580836'),
            pd.Timestamp('2018-01-01 00:00:00.703661761'),
            pd.Timestamp('2018-01-01 00:00:00.801011149'),
            pd.Timestamp('2018-01-01 00:00:00.902080726'),
            pd.Timestamp('2018-01-01 00:00:00.995205845'),
            pd.Timestamp('2018-01-01 00:00:01.104699099'),
        ]
    )
    expected_meta = {'epoch': {'onset': pd.Timestamp('2018-01-01 00:00:00.595580836'),
                               'context': 'foobar',
                               'before': before,
                               'after': after}}
    assert node.o.meta == expected_meta
    pd.testing.assert_frame_equal(node.o.data, expected_data)


def test_empty_epoch():
    # no data received in the epoch
    before = 1
    after = -1
    data = helpers.DummyData(rate=.1)
    node = Epoch(event_trigger='test', before=before, after=after)
    node.clear()
    node.i.data = data.next(5)
    time = pd.Timestamp("2018-01-01 00:00:10.450714306")  # Sync event to second sample
    event = pd.DataFrame([['test', 'foobar']], [time], columns=['label', 'data'])  # Generate a trigger event
    node.i_events.data = event
    node.update()
    expected_meta = {'epoch': {'onset': pd.Timestamp('2018-01-01 00:00:10.450714306'),
                               'context': 'foobar',
                               'before': before,
                               'after': after}
                     }
    assert node.o.meta == expected_meta
    assert node.o.data.empty


def test_to_xarray():
    """ Test the epoch followed by a conversion to xarray.
    """

    graph = {
        'nodes': [
            {
                'id': 'epoch',
                'module': 'timeflux.nodes.epoch',
                'class': 'Epoch',
                'params': {
                    'before': before,
                    'after': after,
                    'event_trigger': 'test'

                }
            },
            {
                'id': 'to_xarray',
                'module': 'timeflux.nodes.epoch',
                'class': 'ToXArray',
                'params': {
                    'reporting': 'error'
                }
            }
        ],
        'edges': [
            {
                'source': 'epoch:*',
                'target': 'to_xarray'
            }

        ]
    }
    # test with data that has no jitter
    data = helpers.DummyData(jitter=0, rate=rate)
    converted_epoch = Branch(graph)
    converted_epoch.set_port('epoch', port_id='i', data=data.next(20))
    event = pd.DataFrame(
        [
            ['test', 'foo'],
            ['test', 'bar'],
        ], [
            pd.Timestamp('2018-01-01 00:00:00.300000'),
            pd.Timestamp('2018-01-01 00:00:00.400000')
        ], columns=['label', 'data'])
    converted_epoch.set_port('epoch', port_id='i_events', data=event)
    converted_epoch.update()

    expected_meta = {'epochs_context': ['foo', 'bar'], 'rate': 10,
                     'epochs_onset': [pd.Timestamp('2018-01-01 00:00:00.300000'),
                                      pd.Timestamp('2018-01-01 00:00:00.400000')]}
    assert converted_epoch.get_port('to_xarray').meta == expected_meta

    expected_data = xr.DataArray(data=np.array([[[0.658783, 0.692277, 0.849196, 0.249668, 0.489425],
                                                 [0.221209, 0.987668, 0.944059, 0.039427, 0.705575],
                                                 [0.925248, 0.180575, 0.567945, 0.915488, 0.033946],
                                                 [0.69742, 0.297349, 0.924396, 0.971058, 0.944266],
                                                 [0.474214, 0.862043, 0.844549, 0.3191, 0.828915],
                                                 [0.037008, 0.59627, 0.230009, 0.120567, 0.076953],
                                                 [0.696289, 0.339875, 0.724767, 0.065356, 0.31529],
                                                 [0.539491, 0.790723, 0.318753, 0.625891, 0.885978],
                                                 [0.615863, 0.232959, 0.024401, 0.870099, 0.021269]],
                                                [[0.221209, 0.987668, 0.944059, 0.039427, 0.705575],
                                                 [0.925248, 0.180575, 0.567945, 0.915488, 0.033946],
                                                 [0.69742, 0.297349, 0.924396, 0.971058, 0.944266],
                                                 [0.474214, 0.862043, 0.844549, 0.3191, 0.828915],
                                                 [0.037008, 0.59627, 0.230009, 0.120567, 0.076953],
                                                 [0.696289, 0.339875, 0.724767, 0.065356, 0.31529],
                                                 [0.539491, 0.790723, 0.318753, 0.625891, 0.885978],
                                                 [0.615863, 0.232959, 0.024401, 0.870099, 0.021269],
                                                 [0.874702, 0.528937, 0.939068, 0.798783, 0.997934]]]),
                                 dims=('epoch', 'time', 'space'),
                                 coords=([0, 1],
                                         pd.TimedeltaIndex(np.arange(-before, after + 1 / rate, 1 / rate), unit='s'),
                                         [0, 1, 2, 3, 4]))
    xr.testing.assert_equal(converted_epoch.get_port('to_xarray').data, expected_data)

    # now test with jittered data, so that the some epoch have invalid number of samples
    data = helpers.DummyData(jitter=.2, rate=rate)
    converted_epoch = Branch(graph)
    converted_epoch.set_port('epoch', port_id='i', data=data.next(20))
    event = pd.DataFrame(
        [
            ['test', 'foo'],
            ['test', 'bar'],
        ], [
            pd.Timestamp('2018-01-01 00:00:00.300000'),
            pd.Timestamp('2018-01-01 00:00:00.400000')
        ], columns=['label', 'data'])
    converted_epoch.set_port('epoch', port_id='i_events', data=event)

    with pytest.raises(WorkerInterrupt):
        converted_epoch.update()



"""Lab streaming layer nodes

    The lab streaming layer provides a set of functions to make instrument data
    accessible in real time within a lab network. From there, streams can be
    picked up by recording programs, viewing programs or custom experiment
    applications that access data streams in real time.

"""

import os
import pandas as pd
import numpy as np
import uuid
from pylsl import (
    StreamInfo,
    StreamOutlet,
    StreamInlet,
    resolve_stream,
    resolve_byprop,
    pylsl,
)
from time import time
from timeflux.core.node import Node


class Send(Node):

    """Send to a LSL stream.

    Attributes:
        i (Port): Default data input, expects DataFrame.

    Args:
        name (string): The name of the stream.
        type (string): The content type of the stream, .
        format (string): The format type for each channel. Currently, only ``double64`` and ``string`` are supported.
        rate (float): The nominal sampling rate. Set to ``0.0`` to indicate a variable sampling rate.
        source (string, None): The unique identifier for the stream. If ``None``, it will be auto-generated.
        config_path (string, None): The path to an LSL config file.

    Example:
        .. literalinclude:: /../examples/lsl.yaml
           :language: yaml

    """

    _dtypes = {"double64": np.number, "string": object}

    def __init__(
        self,
        name,
        type="Signal",
        format="double64",
        rate=0.0,
        source=None,
        config_path=None,
    ):
        if not source:
            source = str(uuid.uuid4())
        self._name = name
        self._type = type
        self._format = format
        self._rate = rate
        self._source = source
        self._outlet = None
        if config_path:
            os.environ["LSLAPICFG"] = config_path

    def update(self):
        if isinstance(self.i.data, pd.core.frame.DataFrame):
            if not self._outlet:
                labels = list(
                    self.i.data.select_dtypes(include=[self._dtypes[self._format]])
                )
                info = StreamInfo(
                    self._name,
                    self._type,
                    len(labels),
                    self._rate,
                    self._format,
                    self._source,
                )
                channels = info.desc().append_child("channels")
                for label in labels:
                    if not isinstance("string", type(label)):
                        label = str(label)
                    channels.append_child("channel").append_child_value("label", label)
                self._outlet = StreamOutlet(info)
            values = self.i.data.select_dtypes(
                include=[self._dtypes[self._format]]
            ).values
            stamps = self.i.data.index.values.astype(np.float64) / 1e9
            for row, stamp in zip(values, stamps):
                self._outlet.push_sample(row, stamp)


class Receive(Node):

    """Receive from a LSL stream.

    Attributes:
        o (Port): Default output, provides DataFrame and meta.

    Args:
        prop (string): The property to look for during stream resolution (e.g., ``name``, ``type``, ``source_id``).
        value (string): The value that the property should have (e.g., ``EEG`` for the type property).
        timeout (float): The resolution timeout, in seconds.
        channels (list, None): Override the channel names. If ``None``, the names defined in the LSL stream will be used.
        max_samples (int): The maximum number of samples to return per call.
        clocksync (bool): Perform automatic clock synchronization.
        dejitter (bool): Remove jitter from timestamps using a smoothing algorithm to the received timestamps.
        monotonize (bool): Force the timestamps to be monotonically ascending. Only makes sense if timestamps are dejittered.
        threadsafe (bool): Same inlet can be read from by multiple threads.
        config_path (string, None): The path to an LSL config file.

    Example:
        .. literalinclude:: /../examples/lsl_multiple.yaml
           :language: yaml

    """

    def __init__(
        self,
        prop="name",
        value=None,
        timeout=1.0,
        channels=None,
        max_samples=1024,
        clocksync=True,
        dejitter=False,
        monotonize=False,
        threadsafe=True,
        config_path=None,
    ):
        if not value:
            raise ValueError("Please specify a stream name or a property and value.")
        self._prop = prop
        self._value = value
        self._inlet = None
        self._labels = None
        self._channels = channels
        self._timeout = timeout
        self._max_samples = max_samples
        self._flags = 0
        self._offset = 0
        if clocksync:
            self._flags |= 1
            self._offset = time() - pylsl.local_clock()
        if dejitter:
            self._flags |= 2
        if monotonize:
            self._flags |= 4
        if threadsafe:
            self._flags |= 8
        if config_path:
            os.environ["LSLAPICFG"] = config_path

    def update(self):
        if not self._inlet:
            self.logger.debug(f"Resolving stream with {self._prop} {self._value}")
            streams = resolve_byprop(self._prop, self._value, timeout=self._timeout)
            if not streams:
                return
            self.logger.debug("Stream acquired")
            self._flags = pylsl.proc_clocksync | pylsl.proc_dejitter
            self._inlet = StreamInlet(streams[0], processing_flags=self._flags)
            info = self._inlet.info()
            self._meta = {
                "name": info.name(),
                "type": info.type(),
                "rate": info.nominal_srate(),
                "info": str(info.as_xml()).replace("\n", "").replace("\t", ""),
            }
            if isinstance(self._channels, list):
                self._labels = self._channels
            else:
                description = info.desc()
                channel = description.child("channels").first_child()
                self._labels = [channel.child_value("label")]
                for _ in range(info.channel_count() - 1):
                    channel = channel.next_sibling()
                    self._labels.append(channel.child_value("label"))
        if self._inlet:
            values, stamps = self._inlet.pull_chunk(max_samples=self._max_samples)
            if stamps:
                if self._offset != 0:
                    self._offset
                    stamps = np.array(stamps) + self._offset
                stamps = pd.to_datetime(stamps, format=None, unit="s")
                self.o.set(values, stamps, self._labels, self._meta)



"""Generate random events"""

import copy
import datetime
import json
import random
import string

import numpy as np
import pandas as pd

from timeflux.core.node import Node
from timeflux.helpers.clock import now
from timeflux.helpers.port import match_events, traverse


class Events(Node):
    def __init__(
        self,
        rows_min=1,
        rows_max=10,
        string_min=3,
        string_max=12,
        items_min=0,
        items_max=5,
        seed=None,
    ):
        """Return random integers from value_min to value_max (inclusive)"""
        self._rows_min = rows_min
        self._rows_max = rows_max
        self._string_min = string_min
        self._string_max = string_max
        self._items_min = items_min
        self._items_max = items_max
        random.seed(seed)

    def random_string(self, length):
        return "".join(random.choice(string.ascii_letters) for m in range(length))

    def update(self):
        rows = []
        for i in range(random.randint(self._rows_min, self._rows_max)):
            row = []
            row.append(
                self.random_string(random.randint(self._string_min, self._string_max))
            )
            data = {}
            for j in range(random.randint(self._items_min, self._items_max)):
                key = self.random_string(
                    random.randint(self._string_min, self._string_max)
                )
                value = self.random_string(
                    random.randint(self._string_min, self._string_max)
                )
                data[key] = value
            row.append(json.dumps(data))
            rows.append(row)
        self.o.set(rows, names=["label", "data"])


class Periodic(Node):
    """Node that sends events at a regular interval.

    This node sends regular events after the first time the update method is
    called. If the update method is called at time `t`, and this node has a
    interval `ti` and phase `ts`, then the first event will be at `t + ts`.
    Then there will be one event at `t + ts + k ti` where `k` is 1, 2, ...

    Args:
        label (str): Event name that will be generated by this node.
        data (dict): Dictionary sent in each event.
        interval (dict): Dictionary with keyword arguments passed to
            `datetime.timedelta` to define the time interval between events.
            This can be seconds, milliseconds, microseconds, etc.
        phase (dict): Dictionary with keyword arguments passed to
            `datetime.timedelta` to define a phase for the stimulations. The
            first stimulation will happen after this time delta is observed.
            If not set, the phase will be as the interval.

    Attributes:
        o (Port): Default output, provides a pandas.DataFrame with events.

    Examples:

        The following YAML can be used to generate events every half second
        but only after 5 seconds have elapsed

        .. code-block:: yaml

           graphs:
              - nodes:
                - id: clock
                  module: timeflux.nodes.events
                  class: Periodic
                  params:
                    label: my-event-label
                    interval:
                      milliseconds: 500
                    phase:
                      seconds: 5

                - id: display
                  module: timeflux.nodes.debug
                  class: Display

                rate: 20

                edges:
                  - source: clock
                    target: display
    """

    def __init__(self, label="clock", data=None, interval=None, phase=None):
        interval = interval or {}
        phase = phase or interval
        delta_interval = datetime.timedelta(**interval)
        delta_phase = datetime.timedelta(**phase)
        if delta_interval.total_seconds() <= 0:
            raise ValueError("Periodic must have positive interval")
        if delta_phase.total_seconds() <= 0:
            raise ValueError("Periodic must have positive phase")

        self._period = np.timedelta64(delta_interval)
        self._phase = np.timedelta64(delta_phase)
        self._next_timestamp = None
        self._label = label
        self._data = data

    def update(self):
        ts = now()
        if self._next_timestamp is None:
            self._next_timestamp = ts + self._phase
            self.logger.debug(
                "Periodic will start sending events at %s", self._next_timestamp
            )

        data = []
        times = []
        while self._next_timestamp < ts:
            content = copy.deepcopy(self._data)
            data.append([self._label, content])
            times.append(self._next_timestamp)
            self._next_timestamp += self._period

        if data:
            self.logger.debug("Sending %d events", len(data))
            self.o.set(data, names=["label", "data"], timestamps=times)

    def _set_next(self, reference):
        self._next_timestamp = reference + self._period


class EventToPulse(Node):

    """Transform an event stream into a pulsed signal stream.

    Attributes:
        i (Port): Default data input, expects DataFrame.

    Args:
        width (float): The pulse width, in seconds.
        amplitude (float): The amplitude of the signal.
        label (string): The name of the marker.

    """

    def __init__(self, width=0.01, amplitude=1, label="marker"):
        self._delta = pd.Timedelta(width, unit="s")
        self._amplitude = amplitude
        self._name = label

    def update(self):
        if self.i.ready():
            index = []
            for timestamp in self.i.data.index:
                index.append(timestamp - self._delta)
                index.append(timestamp)
                index.append(timestamp + self._delta)
            data = [self._amplitude, 0] * len(self.i.data.index)
            self.o.data = pd.DataFrame({self._name: data}, index=index)


class EventToSignal(Node):

    """Transform an event stream into a signal stream.

    Attributes:
        i (Port): Default data input, expects DataFrame.

    Args:
        label (string): The name of the event.
        keys (tuple|str): The hiearchical list of keys. If None, assume scalar. Default: None
    """

    def __init__(self, label="marker", keys=None):
        self._label = label
        self._keys = keys

    def update(self):
        if not self.i.ready():
            return
        events = match_events(self.i, self._label)
        if events is None:
            self.o.data = None
            return
        index = events.index
        if self._keys is None:
            # Assume scalar values
            data = events["data"]
        else:
            data = []
            for value in events["data"]:
                if type(value) == str:
                    # Assume JSON
                    try:
                        value = json.loads(value)
                    except:
                        value = None
                if type(value) == dict:
                    # Get the requested value
                    data.append(traverse(value, self._keys))
                else:
                    # Unhandled type
                    data.append(None)
        self.o.data = pd.DataFrame({self._label: data}, index=index)



"""Tests for events.py"""

import pytest
import logging
import pandas as pd
from timeflux.nodes.events import EventToSignal
from timeflux.helpers.port import make_event

def test_empty():
    node = EventToSignal(label="foobar")
    node.i.data = make_event("marker", 42, serialize=False)
    node.update()
    assert node.o.data == None

def test_scalar():
    node = EventToSignal()
    node.i.data = make_event("marker", 42, serialize=False)
    node.update()
    assert node.o.data["marker"].iloc[0] == 42

def test_dict():
    node = EventToSignal(keys=("meaning", "life"))
    node.i.data = make_event("marker", {"meaning": {"life": 42}}, serialize=False)
    node.update()
    assert node.o.data["marker"].iloc[0] == 42

def test_json():
    node = EventToSignal(keys=("meaning", "life"))
    node.i.data = make_event("marker", {"meaning": {"life": 42}}, serialize=True)
    node.update()
    assert node.o.data["marker"].iloc[0] == 42

def test_invalid_keys():
    node = EventToSignal(keys=("meaning", "life"))
    node.i.data = make_event("marker", {"meaning_of_life":42}, serialize=False)
    node.update()
    assert node.o.data["marker"].iloc[0] == None

def test_invalid_json():
    node = EventToSignal(keys=("meaning", "life"))
    node.i.data = make_event("marker", '{"meaning_of_life":42', serialize=False)
    node.update()
    assert node.o.data["marker"].iloc[0] == None

def test_multiple():
    node = EventToSignal()
    index = [pd.Timestamp('2023-01-01 00:00:00.000'), pd.Timestamp('2023-01-01 00:00:00.100')]
    label = ["marker", "marker"]
    data = [1, 2]
    node.i.data = pd.DataFrame({"label": label, "data": data}, index=index)
    node.update()
    assert len(node.o.data) == 2
    assert node.o.data["marker"].iloc[0] == 1
    assert node.o.data["marker"].iloc[1] == 2
